
#include "Camera.h"
#include "drawutil.h"

#include <GL/glut.h>


Camera* camera = 0;

int timerInterval = 1000;  // in milliseconds
int frameCount = 0;

bool cullBackfaces = true;
bool drawTarget = false;
Ray pickingRay;
bool drawPickingRay = true;
bool drawWorldAxes = true;

enum DrawMode {
   WIREFRAME,
   FLAT_SHADED,
   SMOOTH_SHADED,
   NUM_DRAW_MODES
} drawMode = FLAT_SHADED;

Point3 cubeCentre( 1, 2, 3 );
Point3 sphereCentre( -3, -2, -1 );

bool LMB=false, MMB=false, RMB=false;
bool CTRL=false, ALT=false, SHIFT=false;
int mouse_x, mouse_y;

// MI stands for Menu Item
#define MI_CYCLE_DRAW_MODE 10
#define MI_TOGGLE_DISPLAY_OF_WORLD_SPACE_AXES 21
#define MI_TOGGLE_DISPLAY_OF_CAMERA_TARGET 22
#define MI_TOGGLE_DISPLAY_OF_BACKFACES 23
#define MI_TOGGLE_DISPLAY_OF_PICKING_RAY 24
#define MI_LOOKAT_CUBE 30
#define MI_RESET_CAMERA 40
#define MI_QUIT 50


void reshapeCallback( int width, int height ) {
   if ( 0 != camera )
      camera->resizeViewport( width, height );
}

void drawCallback() {

   Matrix m;
   float s = 1.0f/200;

   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

   glDepthFunc( GL_LEQUAL );
   glEnable( GL_DEPTH_TEST );

   if ( cullBackfaces ) glEnable( GL_CULL_FACE );
   else glDisable( GL_CULL_FACE );

   camera->transform();

   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();

   // ----- draw world space axes

   if ( drawWorldAxes ) {
      drawAxes( Point3(0,0,0), 1, true );

      glColor3f( 0.5f, 0.5f, 0.5f );
      m.setToLookAt( Point3(1,0,0), Point3(0,0,0), Vector3(0,1,0), true );
      glPushMatrix();
         glMultMatrixf( m.get() );
         glScalef( s, s, s );
         glutStrokeCharacter( GLUT_STROKE_MONO_ROMAN, 'x' );
      glPopMatrix();
      m.setToLookAt( Point3(0,1,0), Point3(0,0,0), Vector3(1,0,0), true );
      glPushMatrix();
         glMultMatrixf( m.get() );
         glScalef( s, s, s );
         glutStrokeCharacter( GLUT_STROKE_MONO_ROMAN, 'y' );
      glPopMatrix();
      m.setToLookAt( Point3(0,0,1), Point3(0,0,0), Vector3(0,1,0), true );
      glPushMatrix();
         glMultMatrixf( m.get() );
         glScalef( s, s, s );
         glutStrokeCharacter( GLUT_STROKE_MONO_ROMAN, 'z' );
      glPopMatrix();
   }

   // ----- draw a string suspended in space

   m.setToLookAt( Point3(2,-2,2), Point3(0,0,0), Vector3(0,1,0), true );
   glPushMatrix();
      glMultMatrixf( m.get() );
      drawAxes( Point3(0,0,0), 0.5f, true );
      glScalef( s, s, s );
      char string[] = "Hello";
      glColor3f( 1, 1, 1 );
      for ( int j = 0; string[j] != '\0'; ++j )
         glutStrokeCharacter( GLUT_STROKE_MONO_ROMAN, string[j] );
   glPopMatrix();

   // ----- draw another string, using a utility routine

   char * greeting = "Bonjour";
   float greetingHeight = 2;
   bool useTextureFont = true;
   glColor3f( 0, 1, 0 );
   drawString(
      Point3(
         -widthOfStringToBeDrawn(
            greeting, greetingHeight, DU_FONT_ASCENT, useTextureFont
         )/2,
         0,
         -10
      ),
      greeting, greetingHeight, DU_FONT_ASCENT,
      Vector3(0,0,1), Vector3(0,1,0),
      true, 1, useTextureFont
   );

   // ----- draw a circle

   glColor3f( 1, 0, 0 );
   Point3 circleCentre( 0, 4, 0 );
   drawCircle(
      circleCentre, Vector3( 0, 1, 0 ), 2,
      // The circle is drawn as a polyline.
      // To get a smooth-looking circle, we need about
      // one line segment per pixel.
      camera->convertPixelLength( circleCentre, 1 )
   );

   // ----- setup for drawing objects

   if ( drawMode == WIREFRAME ) {
      glPolygonMode( GL_FRONT_AND_BACK, GL_LINE );
   }
   else {
      Point3 lightPosition = camera->getPosition()
         + ( camera->getPosition() - camera->getTarget() );
      glLightfv( GL_LIGHT0, GL_POSITION, lightPosition.get() );
      glEnable( GL_LIGHT0 );
      glEnable( GL_LIGHTING );

      glShadeModel( drawMode==SMOOTH_SHADED ? GL_SMOOTH : GL_FLAT );
   }

   glColor3f( 1, 1, 1 );

   // ----- draw objects

   glPushMatrix();
      // translate to the cube's center
      glTranslatef( cubeCentre.x(), cubeCentre.y(), cubeCentre.z() );
      glutSolidCube( 1 );
   glPopMatrix();
   glPushMatrix();
      // translate to the sphere's center
      glTranslatef( sphereCentre.x(), sphereCentre.y(), sphereCentre.z() );
      glutSolidSphere( 1, 10, 10 );
   glPopMatrix();

   if ( drawMode == WIREFRAME ) glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
   else glDisable( GL_LIGHTING );

   // ----- draw the selected picking ray

   if ( drawPickingRay ) {
      glColor3f( 0, 1, 1 );
      drawRay( pickingRay );
   }

   // ----- draw little crosshairs at the camera's target

   if ( drawTarget ) {
      const Point3& t = camera->getTarget();
      glColor3f( 0, 1, 1 );
      drawCrossHairs( t, camera->convertLength( t, 0.05 ) );
   }

   // ----- finish up

   glutSwapBuffers();
   ++ frameCount;
}

void mouseCallback( int button, int state, int x, int y ) {
   int mod = glutGetModifiers();
   CTRL = mod & GLUT_ACTIVE_CTRL;
   ALT = mod & GLUT_ACTIVE_ALT;
   SHIFT = mod & GLUT_ACTIVE_SHIFT;

   switch ( button ) {
      case GLUT_LEFT_BUTTON : LMB = state == GLUT_DOWN; break;
      case GLUT_MIDDLE_BUTTON : MMB = state == GLUT_DOWN; break;
      case GLUT_RIGHT_BUTTON : RMB = state == GLUT_DOWN; break;
   }

   if ( state == GLUT_DOWN && !(CTRL || ALT) ) {
      pickingRay = camera->computeRay( x, y );
      glutPostRedisplay();
   }

   mouse_x = x;
   mouse_y = y;

   printf("buttons:(%d,%d,%d); modifiers:(%c,%c,%c); position:(%d,%d)\n",
      LMB ? 1 : 0, MMB ? 1 : 0, RMB ? 1 : 0,
      CTRL ? 'C' : '-', ALT ? 'A' : '-', SHIFT ? 'S' : '-',
      mouse_x, mouse_y
   );
}

void passiveMotionCallback( int x, int y ) {
   mouse_x = x;
   mouse_y = y;
   // printf("move:(%d,%d)\n", mouse_x, mouse_y );
}

void motionCallback( int x, int y ) {
   int delta_x = x - mouse_x;
   int delta_y = mouse_y - y;
   int delta = delta_x + delta_y;

   if ( SHIFT ) {
      // The redundant functionality here is for users with 2-button mice.
      if ( LMB ) {
         if ( ALT )
            camera->dollyCameraForward( 3*delta, false );
         else
            camera->translateSceneRightAndUp( delta_x, delta_y );
      }
   }
   else if ( CTRL ) {
      if ( LMB && MMB )
         camera->rollCameraRight( delta );
      else if ( LMB ) {
         camera->pitchCameraUp( delta_y );
         camera->yawCameraRight( delta_x );
      }
      else if ( MMB )
         camera->zoomIn( delta );
   }
   else if ( ALT ) {
      if ( LMB && MMB )
         camera->dollyCameraForward( 3*delta, false );
      else if ( LMB )
         camera->orbit( mouse_x, mouse_y, x, y );
      else if ( MMB )
         camera->translateSceneRightAndUp( delta_x, delta_y );
   }

   glutPostRedisplay();

   mouse_x = x;
   mouse_y = y;
   // printf("drag:(%d,%d)\n", mouse_x, mouse_y );
}

void menuCallback( int menuItem ) {
   switch ( menuItem ) {
      case MI_CYCLE_DRAW_MODE :
         drawMode = (DrawMode)( ( drawMode + 1 ) % NUM_DRAW_MODES );
         glutPostRedisplay();
         break;
      case MI_TOGGLE_DISPLAY_OF_WORLD_SPACE_AXES :
         drawWorldAxes = ! drawWorldAxes;
         glutPostRedisplay();
         break;
      case MI_TOGGLE_DISPLAY_OF_CAMERA_TARGET :
         drawTarget = ! drawTarget;
         glutPostRedisplay();
         break;
      case MI_TOGGLE_DISPLAY_OF_BACKFACES :
         cullBackfaces = ! cullBackfaces;
         glutPostRedisplay();
         break;
      case MI_TOGGLE_DISPLAY_OF_PICKING_RAY :
         drawPickingRay = ! drawPickingRay;
         glutPostRedisplay();
         break;
      case MI_LOOKAT_CUBE :
         camera->lookAt( cubeCentre );
         glutPostRedisplay();
         break;
      case MI_RESET_CAMERA :
         camera->reset();
         glutPostRedisplay();
         break;
      case MI_QUIT :
         exit(0);
         break;
      default:
         printf("unknown menu item %d was selected\n", menuItem );
         break;
   }
}

void keyboardCallback( unsigned char key, int x, int y ) {
   switch ( key ) {
      case 'a':
         menuCallback( MI_TOGGLE_DISPLAY_OF_WORLD_SPACE_AXES );
         break;
      case 'b':
         menuCallback( MI_TOGGLE_DISPLAY_OF_BACKFACES );
         break;
      case 'd':
         menuCallback( MI_CYCLE_DRAW_MODE );
         break;
      case 'l':
         menuCallback( MI_LOOKAT_CUBE );
         break;
      case 'r':
         menuCallback( MI_RESET_CAMERA );
         break;
      case 't':
         menuCallback( MI_TOGGLE_DISPLAY_OF_CAMERA_TARGET );
         break;
      case 27: // Escape
         menuCallback( MI_QUIT );
         break;
      default:
         printf("untrapped key %d\n", (int)key );
         break;
   }
}

void specialCallback( int key, int x, int y ) {
   switch ( key ) {
      // function keys
      case GLUT_KEY_F1 :
      case GLUT_KEY_F2 :
      case GLUT_KEY_F3 :

      // arrow keys
      case GLUT_KEY_LEFT :
      case GLUT_KEY_RIGHT :
      case GLUT_KEY_UP :
      case GLUT_KEY_DOWN :

      default:
         printf("untrapped special key %d\n", (int)key );
         break;
   }
}

void timerCallback( int /* id */ ) {

   // puts("timer callback called!");

   if ( frameCount > 1 ) {
      printf("%d frames/sec\n",frameCount);
   }
   else if ( frameCount == 1 ) {
      puts("1 or less frames/sec");
   }
   frameCount = 0;

   glutTimerFunc( timerInterval, timerCallback, 0 );
}

int main( int argc, char *argv[] ) {

   glutInit( &argc, argv );
   glutInitDisplayMode( GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE );
   glutInitWindowSize( 512, 512 );
   glutCreateWindow("My Window");

   glutDisplayFunc( drawCallback );
   glutReshapeFunc( reshapeCallback );
   glutKeyboardFunc( keyboardCallback );
   glutSpecialFunc( specialCallback );
   glutMouseFunc( mouseCallback );
   glutPassiveMotionFunc( passiveMotionCallback );
   glutMotionFunc( motionCallback );

   glutTimerFunc( timerInterval, timerCallback, 0 );

   // create a menu
   int mainMenu = glutCreateMenu( menuCallback );
   glutSetMenu( mainMenu ); // make it the current menu
   glutAddMenuEntry( "Cycle Display Mode (d)", MI_CYCLE_DRAW_MODE );
      int displayToggleMenu = glutCreateMenu( menuCallback );
      glutSetMenu( displayToggleMenu );
      glutAddMenuEntry( "World Space Axes (a)",
         MI_TOGGLE_DISPLAY_OF_WORLD_SPACE_AXES );
      glutAddMenuEntry( "Camera Target (t)",
         MI_TOGGLE_DISPLAY_OF_CAMERA_TARGET );
      glutAddMenuEntry( "Backfaces (b)", MI_TOGGLE_DISPLAY_OF_BACKFACES );
      glutAddMenuEntry( "Picking Ray", MI_TOGGLE_DISPLAY_OF_PICKING_RAY );
   glutSetMenu( mainMenu );
   glutAddSubMenu( "Toggle Display of...", displayToggleMenu );
   glutAddMenuEntry( "Look At Cube (l)", MI_LOOKAT_CUBE );
   glutAddMenuEntry( "Reset Camera (r)", MI_RESET_CAMERA );
   glutAddMenuEntry( "Quit (Esc)", MI_QUIT );
   glutAttachMenu( GLUT_RIGHT_BUTTON );//attach the menu to the current window

   int width = glutGet( GLUT_WINDOW_WIDTH );
   int height = glutGet( GLUT_WINDOW_HEIGHT );
   camera = new Camera(
      width, height, 8, Point3( 0, 0, 0)
   );

   glutMainLoop();
}

